/***************************************************************************
 *   Copyright (C) 2005 by Joel R. Voss                                    *
 *   jvoss@myuw.net                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
Robot Cat 1 will be designed to the AltSci ABCD robot specification.

Input A relates to known systems B <-> model C -> output D.

A input (audio, video, proximity)
B known systems (object, person, robot)
C model (love, hate tennis, soccer, etc)
D output (knowledge, action)

A to B: recognition (speech, image)
B to C: understanding (contest)
C to D: decision (purpose)

Robot Cat 1 will use the following technologies to implement ABCD.
recognition: neural networking
understanding: optimized contest
decision: purpose system
*/


#include <stdlib.h>
#include <stdio.h>
#include "robotcat.h"

void RobotCat_init(RobotCat *cat, int inputs, int layers)
{
	// TODO: give the cat an identifier.
	cat->id   = 0;
	// TODO: give the cat a friendly name.
	cat->name = 0;
	cat->current_x = 0;
	cat->current_y = 0;
	// TODO: give the cat a quick reference lookup.
	Memory_init(&(cat->knowledge), 0, 0, start, 0);
	// Initialize the neural network to turn input into known models.
	NeuralNetwork_init(&(cat->recognition), inputs, layers);
	NeuralNetwork_random(&(cat->recognition));
}

void RobotCat_destroy(RobotCat *cat)
{
	cat->id = 0;
	free(cat->name);
	Memory_destroy(&(cat->knowledge));
	NeuralNetwork_destroy(&(cat->recognition));
}

void RobotCat_input(RobotCat *cat, MemType type, BYTE value)
{
	// Create a new memory
	Memory *new_memory = (Memory *)malloc(sizeof(Memory));
	Memory_init(new_memory, cat->current_x, cat->current_y, type, value);
	Memory_add(&(cat->knowledge), new_memory);
}


void RobotCat_explore(RobotCat *cat)
{
	// Put the cat on a mission to seek new stuff.
}

void RobotCat_seek(RobotCat *cat)
{
	// Put the cat on a mission to seek old stuff.
}

void RobotCat_avoid(RobotCat *cat)
{
	// Put the cat on a mission to avoid old stuff.
}

void RobotCat_communicate(RobotCat *cat)
{
	// Put the cat on a mission to communicate worldly knowledge.
}

void RobotCat_update(RobotCat *cat)
{
	// Things that Robot Cat can do:
	
	// TODO: Get input
	// TODO: input should ensure that there are the same number of input 
	// nodes as there are inputs.
	NeuralInput input[] = {true, false, true, false, true, true, false, 
		true, true, false, true};
	
	// Put it into the network.
	NeuralOutput *output = NeuralNetwork_update(&(cat->recognition), input);
	
	// Draw the output;
	printf("Output: ");
	int i;
	for(i = 0; i < cat->recognition.data[0].num_nodes; i++)
	{
		printf("%i ", output[i]);
	}
	printf("\n");
	
	// TODO: Learn from the input.
	
	// No need for memory leaks.
	free(output);
	// Return the output properly?
	return;
}

#define MEMORY_TEST false
#ifndef MEMORY_TEST
#define MEMORY_TEST true
#endif // MEMORY_TEST

void Memory_init(Memory *memory, WORD x, WORD y, MemType type, BYTE value)
{
	memory->position.x = x;
	memory->position.y = y;
	memory->type = type;
	memory->value = value;
	// Do not dare call Memory_init on something that has already been added to!
	memory->next_memory = 0;
	if(MEMORY_TEST) printf("Memory init: %x, val %i, %i B.\n", memory, value, sizeof(Memory));
}

void Memory_add(Memory *memory, Memory *child)
{
	if(memory->next_memory == 0)
	{
		if(MEMORY_TEST) printf("Memory add: %x, %x.\n", memory, child);
		memory->next_memory = child;
	}
	else
	{
		if(MEMORY_TEST) printf("Memory add recurse: %x, %x.\n", memory, child);
		Memory_add(memory->next_memory, child);
	}
}

void Memory_destroy(Memory *memory)
{
	if(memory->next_memory != 0)
	{
		if(MEMORY_TEST) printf("Memory destroy: %x, %x.\n", memory, memory->next_memory);
		Memory_destroy(memory->next_memory);
		free(memory->next_memory);
		memory->next_memory = 0;
	}
}

